#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <ctype.h>

#include "ilgate.h"

#include "auth.h"

void mkkey (char *k);
int speaksfor (char *host, char *user);

int
findkey (char *user, char key[DESKEYLEN])
{
	FILE *f;
	int userlen;
	char buf[100];
	int val;
	char *p;
	int i;

	if ((f = fopen (KEYFILE, "r")) == NULL)
		return (-1);

	userlen = strlen (user);

	while (fgets (buf, sizeof buf, f) != NULL)
		if (strncmp (buf, user, userlen) == 0
		    && isspace (buf[userlen]))
			goto match;

	fclose (f);
	return (-1);

 match:
	fclose (f);

	for (p = buf; *p && !isspace (*p); p++)
		;

	while (isspace (*p))
		p++;

	for (i = 0; i < 7; i++) {
		if (sscanf (p, "%2x", &val) != 1) {
			printf ("%s: syntax error: %s\n", KEYFILE, buf);
			break;
		}
		p += 2;
		key[i] = val;
	}

	return (0);
}

struct authstate {
	struct Ticketreq req;
	char response[NETCHLEN];
	int rcount;
	char hkey[DESKEYLEN], akey[DESKEYLEN];
	unsigned long chal;
};

int
auth_srv (struct task *task, struct ilconn *ilconn)
{
	int i;
	char data[1000];
	int n;
	struct authstate *sp;
	struct Ticket t;

	sp = task_userdata (task, sizeof *sp);

	switch (task_pc (task)) {
	case 0:
		if ((n = il_read (ilconn, data, sizeof data)) < 0) {
			printf ("il_read error\n");
			return (-1);
		}
			
		if (n == 0) /* no data has arrived yet */
			return (0);

		if (n < TICKREQLEN) {
			printf ("auth_srv: bad data len %d\n", n);
			return (-1);
		}

		convM2TR (data, &sp->req);

		if (debug_flag) {
			printf ("auth: ");
			switch (sp->req.type) {
			case AuthTreq: printf ("AuthTreq "); break;
			case AuthChal: printf ("AuthChal "); break;
			case AuthPass: printf ("AuthPass "); break;
			default: printf ("Auth%d ", sp->req.type); break;
			}
			printf ("aid '%s' adom '%s' hostid '%s' uid '%s' ",
				sp->req.authid, sp->req.authdom,
				sp->req.hostid, sp->req.uid);
			printf ("chal ");
			for (i = 0; i < CHALLEN; i++)
				printf ("%02x ", sp->req.chal[i] & 0xff);
			printf ("\n");
			
			printf ("raw data:\n");
			dump (data, n);
		}

		switch (sp->req.type) {
		case AuthTreq:
			printf ("ticket request from %s@%s\n",
				sp->req.authid, sp->req.hostid);

			if (findkey (sp->req.authid, sp->akey) < 0) {
				/* 
				 * make one up so caller doesn't
				 * know it was wrong
				 */
				printf ("%s not found in %s\n",
					sp->req.authid, KEYFILE);
				srandom (microtime ());
				mkkey (sp->akey);
			}

			if (findkey (sp->req.hostid, sp->hkey) < 0) {
				printf ("%s not found in %s\n",
					sp->req.hostid, KEYFILE);
				srandom (microtime ());
				mkkey (sp->hkey);
			}
			
			memset (&t, 0, sizeof t);
			memmove (t.chal, sp->req.chal, CHALLEN);
			strcpy (t.cuid, sp->req.uid);
			if (speaksfor (sp->req.hostid, sp->req.uid))
				strcpy (t.suid, sp->req.uid);
			else
				strcpy (t.suid, "none");
			/* translate goes here */

			srandom (microtime () ^ sp->hkey[0] ^ sp->akey[0]);
			mkkey (t.key);

			data[0] = AuthOK;
			t.num = AuthTc;
			convT2M (&t, data + 1, sp->hkey);
			t.num = AuthTs;
			convT2M (&t, data + 1 + TICKETLEN, sp->akey);
			il_write (ilconn, data, 2*TICKETLEN+1);
			task_set_pc (task, 2);
			return (0);

		case AuthChal:
			printf ("challenge request for %s / %s\n",
				sp->req.uid, sp->req.hostid);

			if (findkey (sp->req.uid, sp->akey) < 0) {
				srandom (microtime ());
				mkkey (sp->akey);
			}

			if (findkey (sp->req.hostid, sp->hkey) < 0) {
				srandom (microtime ());
				mkkey (sp->hkey);
			}

			data[0] = AuthOK;
			srandom (microtime () ^ sp->akey[0]);
			sp->chal = random () % MAXNETCHAL;
			sprintf (data + 1, "%lu", sp->chal);
			il_write (ilconn, data, NETCHLEN + 1);
			task_set_pc (task, 1);
			return (0);
		default:
			printf ("unknown request type\n");
			return (-1);
		}
	case 1:
		if ((n = il_read (ilconn,
				  sp->response + sp->rcount,
				  NETCHLEN - sp->rcount)) < 0) {
			printf ("read error\n");
			return (-1);
		}
			
		if (n == 0)
			return (0);

		sp->rcount += n;

		if (sp->rcount < NETCHLEN)
			return (0);

		if (debug_flag) {
			printf ("response:\n");
			dump (sp->response, NETCHLEN);
		}
		
		if (netcheck (sp->akey, sp->chal, sp->response) == 0) {
			printf ("incorrect response from %s / %s\n",
				sp->req.uid, sp->req.hostid);
			memset (data, 0, ERRLEN + 1);
			data[0] = AuthErr;
			sprintf (data + 1, "bad response");
			il_write (ilconn, data, ERRLEN + 1);
			il_close (ilconn);
			task_exit (task);
			return (0);
		} else {
			struct Authenticator a;

			printf ("corrrect response from %s / %s\n",
				sp->req.uid, sp->req.hostid);

			memset (&t, 0, sizeof t);
			memcpy (t.chal, sp->req.chal, CHALLEN);
			strcpy (t.cuid, sp->req.uid);
			strcpy (t.suid, sp->req.uid);
			t.num = AuthTs;

			srandom (microtime () ^ sp->hkey[0]);
			mkkey (t.key);
			data[0] = AuthOK;
			convT2M (&t, data + 1, sp->hkey);
			memcpy (a.chal, t.chal, CHALLEN);
			a.num = AuthAc;
			a.id = 0;
			convA2M (&a, data + TICKETLEN + 1, t.key);
			if (debug_flag) {
				printf ("reply:\n");
				dump (data, TICKETLEN + AUTHENTLEN + 1);
			}
			il_write (ilconn, data, TICKETLEN + AUTHENTLEN + 1);
			task_set_pc (task, 2);
			return (0);
		}
		return (0);
	case 2:
		return (0);

	default:
		printf ("unknown srv pc\n");
		return (-1);
	}
}

void
mkkey (char *k)
{
	int i;
	for (i = 0; i < DESKEYLEN; i++)
		k[i] = random ();
}

int
speaksfor (char *host, char *user)
{
	return (1);
}
